#
# pushare - P2P file sharing system
#
# Copyright (c) 2004 Satoshi Fukutomi <fuktommy@inter7.jp>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. Neither the name of the University nor the names of its contributors
#    may be used to endorse or promote products derived from this software
#    without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
# $Id: NodeList.pm,v 1.3 2004/12/17 15:56:48 fuktommy Exp $
#

package PuShare::NodeList;
use strict;
use PuShare::Config;
use PuShare::Node;
use PuShare::Util;

#
# configration
#
my $nodelist = $PuShare::Config::nodelist;

#
# show all nodes
#
sub list() {
	local $_;
	open IN, $nodelist or return ();
	my @node = ();
	while (<IN>) {
		chomp;
		my ($host, $port) = split /:/;
		my $n = new PuShare::Node($host, $port);
		push @node, $n;
	}
	return @node;
}

#
# show randmized all nodes
#
sub all() {
	local $_;
	my @node = list();
	my $i;
	for ($i=@node; $i>0; $i--) {
		my $t = int rand $i;
		($node[$i-1], $node[$t]) = ($node[$t], $node[$i-1]);
	}
	return @node;
}

#
# suggest random node
#
sub random() {
	local $_;
	my @node = list();
	if (@node) {
		my $node = $node[rand(0+@node)];
		return $node;
	} else {
		return undef;
	}
}

#
# who am i
#
sub myself() {
	my $port = $main::port;
	local $_;
	my @node = all();
	foreach (@node) {
		my @buf = $_->talk("/ping");
		if ((defined $buf[1])&&($buf[0] eq "PONG")) {
			return new PuShare::Node($buf[1], $port);
		}
	}
	return undef;
}

#
# add node to list
# args: node
#
sub add($) {
	my($node) = @_;
	my $str = $node->toString() . "\n";
	local $_;
	unless (include($node)) {
		PuShare::Util::lock();
		open OUT, ">> $nodelist" or die "$nodelist: open failed. $!";
		print OUT "$str";
		close OUT;
		PuShare::Util::unlock();
	}
}

#
# include node
# args: node
#
sub include($) {
	my($node) = @_;
	my $str = $node->toString();
	local $_;
	my @node = list();
	if (grep $_->toString() eq $str, @node) {
		return 1;
	} else {
		return undef;
	}
}

#
# count node
#
sub count() {
	my @node = list();
	return 0+@node;
}

#
# remove node from list
# args: node
#
sub remove($) {
	my($node) = @_;
	my $str = $node->toString();
	$str .= "\n";
	local $_;
	my @node;
	if (-e $nodelist) {
		open IN, $nodelist or exit;
		@node = <IN>;
		close IN;
	} else {
		return;
	}
	if (grep $_ eq "$str", @node) {
		PuShare::Util::lock();
		open OUT, "> $nodelist" or die "$nodelist: open failed. $!";
		foreach (@node) {
			print OUT if ($_ ne $str);
		}
		close OUT;
		PuShare::Util::unlock();
	}
}

#
# ping all nodes
#
sub pingall() {
	local $_;
	my @node = all();
	foreach (@node) {
		unless ($_->ping()) {
			remove($_);
		}
	}
}

#
# join network
# args: node, port
#
sub join($) {
	my ($node) = @_;
	my $port = $main::port;
	my $count = 0;
	while (1) {
		my @buf = $node->join();
		my $n = $node->toString();
		if ((@buf == 1)&&($buf[0] eq "WELCOME")) {
			add($node);
			last;
		} elsif ($count++ > $PuShare::Config::retrySearch) {
			last;
		} elsif ((@buf == 2)&&($buf[0] eq "WELCOME")) {
			add($node);
			my($h, $p) = split /:/, $buf[1], 2;
			$node = new PuShare::Node($n, $p);
		} else {
			last;
		}
	}
}

#
# connect initial node
# args: port
#
sub init() {
	my $port = $ENV{SERVER_PORT};
	my @initNode = @PuShare::Config::initNode;
	unshift @initNode, $main::initNode if ($main::initNode);
	local $_;
	my $inode;
	foreach (@initNode) {
		my($h, $p) = split /:/, $_, 2;
		$inode = new PuShare::Node($h, $p);
		if ($inode->ping()) {
			PuShare::NodeList::join($inode);
			last;
		}
	}
	my $myself = myself();
	remove($myself) if ($myself);
	my @node = all();
	my %done = ();
	return if (@node == 0);

	$done{$_->toString()} = 1 foreach (@node);
	while (1) {
		my $node = $node[rand(0+@node)];
		my @buf  = $node->talk("/node");
		my $str = "";
		if ((@buf > 0)&&($buf[0] =~ m|:\d+/[^: ]+$|)) {
			$str = $buf[0];
			my($h, $p) = split /:/, $str, 2;
			$node = new PuShare::Node($h, $p);
			push @node, $node unless ($done{$str});
		}
		if (($done{$str}++ > $PuShare::Config::retrySearch)||
				(@node >= $PuShare::Config::nodes)) {
			last;
		}
	}

	foreach (@node) {
		unless (include($_)) {
			my($h, $p) = split /:/, $_, 2;
			PuShare::NodeList::join($_);
		}
	}
	remove($myself) if ($myself);
	if (count() >= $PuShare::Config::nodes) {
		$inode->bye();
		remove($inode);
	}
	while (count() > $PuShare::Config::nodes) {
		my $node = random();
		$node->bye();
		remove($node);
	}
}

#
# relay distribution
#
sub relayDist($) {
	my($message) = @_;
	my $pid;
	if (rand() < $PuShare::Config::ignore) {
		return;
	} elsif (($pid = fork) || (! defined $pid)) {
		$SIG{CHLD} = "IGNORE";
		return;
	}
	my @node = list();
	local $_;
	foreach (@node) {
		$_->talk("/dist?$message");
	}
	exit;
}

1;
