#!/usr/bin/perl -w
#
# pushare - P2P file sharing system
#
# Copyright (c) 2004 Satoshi Fukutomi <fuktommy@inter7.jp>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. Neither the name of the University nor the names of its contributors
#    may be used to endorse or promote products derived from this software
#    without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
# $Id: pushare.pl,v 1.6 2004/12/17 15:56:48 fuktommy Exp $
#

use strict;
use lib qw(.);
use Socket;
use PuShare::Cache;
use PuShare::Config;
use PuShare::Interface;
use PuShare::Gateway;
use PuShare::Node;
use PuShare::NodeList;
use PuShare::Search;

$SIG{CHLD} = "IGNORE";
use vars qw(	$port $initNode 
		$path $query_string $remote_addr);

#
# default value
#
$port = $PuShare::Config::port;
my $timeout = $PuShare::Config::timeout;
my $cachedir = $PuShare::Config::cachedir;
my $bufsize  = $PuShare::Config::bufsize;
my($pid_client);


while (@ARGV) {
	$_ = shift @ARGV;
	if ($_ eq "-p") {
		$port = shift @ARGV;
	} elsif ($_ eq "-i") {
		$initNode = shift @ARGV;
	}
}



#------------------------------------------------------------------#

sub xdie(@) {
	if ((defined $pid_client)&&($pid_client>0)) {
		kill "TERM", $pid_client;
		dprint "terminate $pid_client";
	}
	die @_;
}

sub dist_include($) {
	my($id) = @_;
	local $_;
	if (open IN, $PuShare::Config::distlist) {
		while (<IN>) {
			if (/$id/) {
				close IN;
				return 1;
			}
		}
		close IN;
	}
	return 0;
}

sub xlocaltime(;$) {
	my @buf = (@_)? localtime(@_): localtime;
	return sprintf "%d-%02d-%02d %02d:%02d:%02d",
		       1900+$buf[5], $buf[4]+1, $buf[3], $buf[2], $buf[1], $buf[0];
}

#------------------------------------------------------------------#
#
# httpd
#
sub httpd_main() {
	$SIG{TERM} = $SIG{HUP} = "DEFAULT";
	$SIG{CHLD} = "IGNORE";
	$SIG{ALRM} = sub { die "$remote_addr: server timeout" };

	# parse HTTP request
	my($str_request, $http_user_agent) = ("", "");
	alarm $timeout;
	my($request_method, $request_uri, $http_host, $range);

	while (<CLIENT>) {
		s/[\n|\r]//g;
		if ($_ eq "") {
			last;
		} elsif (/^(\S+)\s+(\S+)/ && (!defined $request_method)) {
			$str_request    = $_;
			$request_method = $1;
			$request_uri	= $2;
			$request_uri	=~ s|^//+|/|g;

		} elsif (/^Host:\s+/i) {
			$http_host = $';

		} elsif (/^Range:\s*bytes\s*=\s*(\d+)\s*-/) {
			$range = $1;

		} elsif (/^User-Agent:\s+/i) {
			$http_user_agent = $';

		}
	}
	alarm 0;

	if ($request_uri =~ m|^[a-z]+://([^/]+)|) {
		$http_host   = $1;
		$request_uri = $';
	}
	if ($request_uri eq "") {
		$request_uri = "/";
	}
	if ($request_uri eq "/") {
		$path = "/";
		$query_string = "";
	} elsif ($request_uri =~ /\?/) {
		$path = $`;
		$query_string = $';
	} else {
		$path = $request_uri;
		$query_string = "";
	}

	# output logs
	my $date = xlocaltime();
	dprint "$remote_addr<>$date<>$str_request<>$http_user_agent";

	if ($request_method eq "OPTIONS") {
		printHeader("200 OK");
		print "Allow: GET, HEAD, OPTIONS\n",
		      "Content-Length: 0\r\n\r\n";

	} elsif ($request_method eq "HEAD") {
		printHeader("200 OK");
		print "\r\n";

	} elsif ($request_method !~ /^(GET|HEAD)$/) {
		printHeader("501 Not Implemented");
		print "\r\n";

	} elsif ($path eq "/ping") {
		printHeader("200 OK");
		print "Content-Type: text/plain\r\n",
		      "\r\n",
		      "PONG\n",
		      "$remote_addr\n";

	} elsif ($path eq "/node") {
		my $node = PuShare::NodeList::random();
		if ($node) {
			printHeader("200 OK");
			print "Content-Type: text/plain\r\n",
			      "\r\n",
			      $node = $node->toString(), "\n";
		} else {
			print404();
			return;
		}

	} elsif ($path =~ m|^/get/([0-9a-f]{32})$|) {
		my $file = $1;
		unless ((-f "$cachedir/$file.dat")
			&& (dist_include($file))
			&& (open IN, "$cachedir/$file.dat")) {
			print404();
			return;
		}
		my $size = (-s "$cachedir/$file.dat");
		if (defined $range) {
			my $all = $size;
			$size -= $range;
			seek IN, $range, 00;
			printHeader("206 Partial content");
			print "Content-Range: bytes $range-" . ($all-1)  . "/$all\r\n";
		} else {
			printHeader("200 OK");
		}
		print "Content-Length: $size\r\n",
		      "Content-Type: application/octet-stream\r\n",
		      "\r\n";
		my $buf;
		alarm $timeout;
		while (read IN, $buf, $bufsize) {
			print $buf;
			alarm $timeout;
		}

	} elsif ($path eq "/join") {
		my %arg = args($query_string);
		unless ($arg{port}) {
			print404();
			return;
		}
		my $node = new PuShare::Node($remote_addr, $arg{port});
		unless ($node->ping()) {
			print404();
			return;
		} elsif (PuShare::NodeList::include($node) ||
			(PuShare::NodeList::count() < $PuShare::Config::nodes)) {
			PuShare::NodeList::add($node);
			printHeader("200 OK");
			print "Content-Type: text/plain\r\n",
			      "\r\n",
			      "WELCOME\n";
		} else {
			my $suggest = PuShare::NodeList::random();
			PuShare::NodeList::remove($suggest);
			$suggest->bye();
			printHeader("200 OK");
			print "Content-Type: text/plain\r\n",
			      "\r\n",
			      "WELCOME\n",
			      $suggest->toString(), "\n";
			PuShare::NodeList::add($node);
		}

	} elsif ($path eq "/bye") {
		my %arg = args($query_string);
		unless ($arg{port}) {
			print404();
			return;
		}
		my $node = new PuShare::Node::($remote_addr, $arg{port});
		PuShare::NodeList::remove($node);
		printHeader("200 OK");
		print "Content-Type: text/plain\r\n",
		      "\r\n",
		      "BYEBYE\n";

	} elsif ($path eq "/dist") {
		my $pid;
		if (($pid = fork) || (! defined $pid)) {
			printHeader("200 OK");
			print "Content-Type: text/plain\r\n",
			      "\r\n",
			      "OK\n";
			return;
		}
		local $_;
		my %arg = args($query_string);
		unless ($arg{id} && $arg{port}) {
			print404();
			return;
		}
		unless ($arg{host}) {
			$arg{host} = $remote_addr;
		}
		my $date = time;

		my @ignore = ();
		if (open IN, $PuShare::Config::ignorelist) {
			@ignore = <IN>;
			close IN;
		}
		if (dist_include($arg{id})) {

		} elsif (-s "$cachedir/$arg{id}.dat") {
			PuShare::Util::addDist($date, $arg{id});
			$query_string =~ s/host=$arg{host}//;
			$query_string =~ s/port=$arg{port}/port=$port/;
			PuShare::NodeList::relayDist($query_string);

		} elsif (grep $_ eq $arg{id}, @ignore) {
			PuShare::Util::addDist($date, $arg{id});
			PuShare::NodeList::relayDist($query_string);

		} elsif (PuShare::Search::check(%arg)) {
			my $id = $arg{id};
			PuShare::Util::addDist($date, $id);
			my $node = new PuShare::Node($remote_addr, $arg{port});
			if ($node->talk("/get/$id", file=>$id)) {
				my $name = decode($arg{name});
				$query_string =~ s/host=$arg{host}//;
				$query_string =~ s/port=$arg{port}/port=$port/;
				PuShare::Util::addIgnore($id);
				PuShare::Cache::add($id, $name);
				PuShare::Cache::convert($id);
				PuShare::NodeList::relayDist($query_string);
			} else {
				PuShare::Util::delDist($date, $id);
			}

		} else {
			PuShare::Util::addDist($date, $arg{id});
			PuShare::NodeList::relayDist($query_string);
		}

	} elsif (($path eq "/") || ($path =~ m|^/gw|)) {
		PuShare::Gateway::main();

	} else {
		print404();
	}
}

sub httpd() {
	# initialize daemon
	socket CLIENT_WAITING, PF_INET, SOCK_STREAM, 0 or xdie "failed to open socket. $!\n";
	setsockopt CLIENT_WAITING, SOL_SOCKET, SO_REUSEADDR, 1 or xdie "failed to setsockopt. $!\n";
	bind CLIENT_WAITING, pack_sockaddr_in($port, INADDR_ANY) or xdie "failed to bind. $!\n";
	listen CLIENT_WAITING, SOMAXCONN or xdie "failed to listen: $!\n";

	my $children = 0;
	my %connection = ();
	my %host = ();
	$SIG{CHLD} = sub {
		my $pid  = wait;
		foreach $pid (keys %host) {
			unless (kill 0, $pid) {
				if (--$connection{$host{$pid}} <= 0) {
					delete $connection{$host{$pid}};
				}
				delete $host{$pid};
				$children--;
			}
		}
		if ($children <= 0) {
			%connection = ();
			%host = ();
			$children = 0;
		}
	};

	# run daemon
	while (1) {
		my $paddr;
		sleep 1 if (rand() < 1/$PuShare::Config::children);
		unless ($paddr = accept(CLIENT, CLIENT_WAITING)) {
			sleep 1;
			next;
		}
		my ($client_port, $client_iaddr) = unpack_sockaddr_in $paddr;
		$remote_addr = inet_ntoa $client_iaddr;
		select CLIENT; $|=1;

		my $pid;
		if (! defined($pid = fork)) {
			dprint "server process: fork failed. $!";
			close CLIENT;
			next;

		} elsif ($pid) {
			close CLIENT;
			$connection{$remote_addr}++;
			$host{$pid} = $remote_addr;
			$children++;
			next;

		} else {
			if ($connection{$remote_addr}) {
				sleep 1;
				dprint "too many connections for the host: $remote_addr";
				printError("503 Service Unavailable");
			} elsif ($children > $PuShare::Config::children) {
				sleep 1;
				dprint "too many connection: $remote_addr";
				printError("503 Service Unavailable");
			} else {
				httpd_main();
			}
			shutdown CLIENT, 2;
			close CLIENT;
			exit;
		}
	}
}

#------------------------------------------------------------------#
#
# client process
#
sub client_save(%) {
	my(%time) = @_;
	local $, = "\n";
	PuShare::Util::lock();
	open OUT, "> $PuShare::Config::client"
		or die "Failed to open $PuShare::Config::client. $!";
	print OUT $time{ping}, $time{join}, $time{init}, $time{dist}, "";
	close OUT;
	PuShare::Util::unlock();
}

sub client_main() {
	$SIG{ALRM} = sub { die "client timeout" };
	alarm $PuShare::Config::timeout_client * 60;
	my %time = ();
	PuShare::Cache::update();
	if (open IN, $PuShare::Config::client) {
		$time{ping} = <IN>;
		$time{join} = <IN>;
		$time{init} = <IN>;
		$time{dist} = <IN>;
		chomp($time{ping}, $time{join}, $time{init}, $time{dist});
		close IN;
	} else {
		($time{ping}, $time{join}, $time{init}, $time{dist}) = (0, 0, 0, 0);
	}
	if ((time - $time{ping}) >= $PuShare::Config::pingfrec * 60) {
		$time{ping} = time;
		client_save(%time);
		PuShare::NodeList::pingall();
		dprint "ping() finished";
	}
	my $nc = PuShare::NodeList::count();
	if ((time - $time{init}) >= $PuShare::Config::initfrec * $nc * 60) {
		$time{init} = time;
		client_save(%time);
		PuShare::NodeList::init();
		dprint "init() finished";
	}
	if ((time - $time{join}) >= $PuShare::Config::joinfrec * 60) {
		$time{join} = time;
		client_save(%time);
		my @node = PuShare::NodeList::all();
		foreach (@node) {
			my $n = $_->toString();
			PuShare::NodeList::join($_);
		}
		dprint "join() finished";
	}
	if ((time - $time{dist}) >= $PuShare::Config::distfrec * 60) {
		$time{dist} = time;
		client_save(%time);
		my @file = glob("$cachedir/*.dat");
		return unless (@file);
		my $i;
		for ($i=$#file; $i>=0; $i--) {
			my $t = int rand $i;
			($file[$i], $file[$t]) = ($file[$t], $file[$i]);
		}
		my $file;
		foreach $file (@file) {
			my $id	 = $file;  $id =~ s|$cachedir/||;  $id =~ s|\.dat$||;
			next if (dist_include($id));
			my $size = (-s $file);
			my $name = PuShare::Cache::name($id);
			if ($name) {
				$name = encode($name);
			} else {
				$name = "";
			}
			PuShare::Util::addDist(time, $id);
			PuShare::NodeList::relayDist("port=$port&id=$id&name=$name");
			last;
		}
	}
}

sub client() {
	$SIG{TERM} = $SIG{HUP} = "DEFAULT";
	$SIG{CHLD} = "IGNORE";
	mkdir $PuShare::Config::updir, 0777 unless (-d $PuShare::Config::updir);
	mkdir $PuShare::Config::downdir, 0777 unless (-d $PuShare::Config::downdir);
	mkdir $PuShare::Config::cachedir, 0777 unless (-d $PuShare::Config::cachedir);
	unlink $PuShare::Config::client;
	sleep 5;

	while (1) {
		my $pid;
		if (! defined ($pid = fork)) {
			dprint "client process: fork failed.";
			sleep $PuShare::Config::clientfrec * 60;
			next;

		} elsif ($pid) {
			sleep 10;
			my $t = $PuShare::Config::clientfrec * 60
			      + int($PuShare::Config::timeError * PuShare::Util::normal());
			$t = $PuShare::Config::clientfrec * 60 if ($t <= 0);
			sleep $t;
			next;

		} else {
			client_main();
			exit;
		}
	}
}


#------------------------------------------------------------------#

PuShare::Util::standby();
if (($pid_client=fork) == 0) {
	client();
}
httpd();
